@@ -1,9 +1,8 @@
-from __future__ import unicode_literals
+import uuid
 
-import django
-from django.db import models
 from django.conf import settings
-from django.utils.encoding import python_2_unicode_compatible
+from django.core.exceptions import ValidationError
+from django.db import models
 from django.utils.timezone import now
 
 DEFAULT_CHOICES_NAME = 'STATUS'
@@ -17,10 +16,11 @@ class AutoCreatedField(models.DateTimeField):
     By default, sets editable=False, default=datetime.now.
 
     """
+
     def __init__(self, *args, **kwargs):
         kwargs.setdefault('editable', False)
         kwargs.setdefault('default', now)
-        super(AutoCreatedField, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
 
 class AutoLastModifiedField(AutoCreatedField):
@@ -30,8 +30,26 @@ class AutoLastModifiedField(AutoCreatedField):
     By default, sets editable=False and default=datetime.now.
 
     """
+    def get_default(self):
+        """Return the default value for this field."""
+        if not hasattr(self, "_default"):
+            self._default = self._get_default()
+        return self._default
+
     def pre_save(self, model_instance, add):
         value = now()
+        if add:
+            current_value = getattr(model_instance, self.attname, self.get_default())
+            if current_value != self.get_default():
+                # when creating an instance and the modified date is set
+                # don't change the value, assume the developer wants that
+                # control.
+                value = getattr(model_instance, self.attname)
+            else:
+                for field in model_instance._meta.get_fields():
+                    if isinstance(field, AutoCreatedField):
+                        value = getattr(model_instance, field.name)
+                        break
         setattr(model_instance, self.attname, value)
         return value
 
@@ -48,20 +66,19 @@ class StatusField(models.CharField):
     Also features a ``no_check_for_status`` argument to make sure
     South can handle this field when it freezes a model.
     """
+
     def __init__(self, *args, **kwargs):
         kwargs.setdefault('max_length', 100)
         self.check_for_status = not kwargs.pop('no_check_for_status', False)
         self.choices_name = kwargs.pop('choices_name', DEFAULT_CHOICES_NAME)
-        super(StatusField, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def prepare_class(self, sender, **kwargs):
         if not sender._meta.abstract and self.check_for_status:
             assert hasattr(sender, self.choices_name), \
                 "To use StatusField, the model '%s' must have a %s choices class attribute." \
                 % (sender.__name__, self.choices_name)
-            self._choices = getattr(sender, self.choices_name)
-            if django.VERSION >= (1, 9, 0):
-                self.choices = self._choices
+            self.choices = getattr(sender, self.choices_name)
             if not self.has_default():
                 self.default = tuple(getattr(sender, self.choices_name))[0][0]  # set first as default
 
@@ -70,13 +87,11 @@ class StatusField(models.CharField):
         # we don't set the real choices until class_prepared (so we can rely on
         # the STATUS class attr being available), but we need to set some dummy
         # choices now so the super method will add the get_FOO_display method
-        self._choices = [(0, 'dummy')]
-        if django.VERSION >= (1, 9, 0):
-            self.choices = self._choices
-        super(StatusField, self).contribute_to_class(cls, name)
+        self.choices = [(0, 'dummy')]
+        super().contribute_to_class(cls, name)
 
     def deconstruct(self):
-        name, path, args, kwargs = super(StatusField, self).deconstruct()
+        name, path, args, kwargs = super().deconstruct()
         kwargs['no_check_for_status'] = True
         return name, path, args, kwargs
 
@@ -88,6 +103,7 @@ class MonitorField(models.DateTimeField):
     changes.
 
     """
+
     def __init__(self, *args, **kwargs):
         kwargs.setdefault('default', now)
         monitor = kwargs.pop('monitor', None)
@@ -99,22 +115,21 @@ class MonitorField(models.DateTimeField):
         if when is not None:
             when = set(when)
         self.when = when
-        super(MonitorField, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def contribute_to_class(self, cls, name):
         self.monitor_attname = '_monitor_%s' % name
         models.signals.post_init.connect(self._save_initial, sender=cls)
-        super(MonitorField, self).contribute_to_class(cls, name)
+        super().contribute_to_class(cls, name)
 
     def get_monitored_value(self, instance):
         return getattr(instance, self.monitor)
 
     def _save_initial(self, sender, instance, **kwargs):
-        if django.VERSION >= (1, 10) and self.monitor in instance.get_deferred_fields():
+        if self.monitor in instance.get_deferred_fields():
             # Fix related to issue #241 to avoid recursive error on double monitor fields
             return
-        setattr(instance, self.monitor_attname,
-                self.get_monitored_value(instance))
+        setattr(instance, self.monitor_attname, self.get_monitored_value(instance))
 
     def pre_save(self, model_instance, add):
         value = now()
@@ -124,10 +139,10 @@ class MonitorField(models.DateTimeField):
             if self.when is None or current in self.when:
                 setattr(model_instance, self.attname, value)
                 self._save_initial(model_instance.__class__, model_instance)
-        return super(MonitorField, self).pre_save(model_instance, add)
+        return super().pre_save(model_instance, add)
 
     def deconstruct(self):
-        name, path, args, kwargs = super(MonitorField, self).deconstruct()
+        name, path, args, kwargs = super().deconstruct()
         kwargs['monitor'] = self.monitor
         if self.when is not None:
             kwargs['when'] = self.when
@@ -139,7 +154,10 @@ SPLIT_MARKER = getattr(settings, 'SPLIT_MARKER', '<!-- split -->')
 # the number of paragraphs after which to split if no marker
 SPLIT_DEFAULT_PARAGRAPHS = getattr(settings, 'SPLIT_DEFAULT_PARAGRAPHS', 2)
 
-_excerpt_field_name = lambda name: '_%s_excerpt' % name
+
+def _excerpt_field_name(name):
+    return '_%s_excerpt' % name
+
 
 def get_excerpt(content):
     excerpt = []
@@ -156,8 +174,8 @@ def get_excerpt(content):
 
     return '\n'.join(default_excerpt)
 
-@python_2_unicode_compatible
-class SplitText(object):
+
+class SplitText:
     def __init__(self, instance, field_name, excerpt_field_name):
         # instead of storing actual values store a reference to the instance
         # along with field names, this makes assignment possible
@@ -166,11 +184,13 @@ class SplitText(object):
         self.excerpt_field_name = excerpt_field_name
 
     # content is read/write
-    def _get_content(self):
+    @property
+    def content(self):
         return self.instance.__dict__[self.field_name]
-    def _set_content(self, val):
+
+    @content.setter
+    def content(self, val):
         setattr(self.instance, self.field_name, val)
-    content = property(_get_content, _set_content)
 
     # excerpt is a read only property
     def _get_excerpt(self):
@@ -185,7 +205,8 @@ class SplitText(object):
     def __str__(self):
         return self.content
 
-class SplitDescriptor(object):
+
+class SplitDescriptor:
     def __init__(self, field):
         self.field = field
         self.excerpt_field_name = _excerpt_field_name(self.field.name)
@@ -205,6 +226,7 @@ class SplitDescriptor(object):
         else:
             obj.__dict__[self.field.name] = value
 
+
 class SplitField(models.TextField):
     def __init__(self, *args, **kwargs):
         # for South FakeORM compatibility: the frozen version of a
@@ -212,17 +234,17 @@ class SplitField(models.TextField):
         # _excerpt field itself is frozen as well. See introspection
         # rules below.
         self.add_excerpt_field = not kwargs.pop('no_excerpt_field', False)
-        super(SplitField, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def contribute_to_class(self, cls, name):
         if self.add_excerpt_field and not cls._meta.abstract:
             excerpt_field = models.TextField(editable=False)
             cls.add_to_class(_excerpt_field_name(name), excerpt_field)
-        super(SplitField, self).contribute_to_class(cls, name)
+        super().contribute_to_class(cls, name)
         setattr(cls, self.name, SplitDescriptor(self))
 
     def pre_save(self, model_instance, add):
-        value = super(SplitField, self).pre_save(model_instance, add)
+        value = super().pre_save(model_instance, add)
         excerpt = get_excerpt(value.content)
         setattr(model_instance, _excerpt_field_name(self.attname), excerpt)
         return value.content
@@ -238,6 +260,52 @@ class SplitField(models.TextField):
             return value
 
     def deconstruct(self):
-        name, path, args, kwargs = super(SplitField, self).deconstruct()
+        name, path, args, kwargs = super().deconstruct()
         kwargs['no_excerpt_field'] = True
         return name, path, args, kwargs
+
+
+class UUIDField(models.UUIDField):
+    """
+    A field for storing universally unique identifiers. Use Python UUID class.
+    """
+
+    def __init__(self, primary_key=True, version=4, editable=False, *args, **kwargs):
+        """
+        Parameters
+        ----------
+        primary_key : bool
+            If True, this field is the primary key for the model.
+        version : int
+            An integer that set default UUID version.
+        editable : bool
+            If False, the field will not be displayed in the admin or any other ModelForm,
+            default is false.
+
+        Raises
+        ------
+        ValidationError
+            UUID version 2 is not supported.
+        """
+
+        if version == 2:
+            raise ValidationError(
+                'UUID version 2 is not supported.')
+
+        if version < 1 or version > 5:
+            raise ValidationError(
+                'UUID version is not valid.')
+
+        if version == 1:
+            default = uuid.uuid1
+        elif version == 3:
+            default = uuid.uuid3
+        elif version == 4:
+            default = uuid.uuid4
+        elif version == 5:
+            default = uuid.uuid5
+
+        kwargs.setdefault('primary_key', primary_key)
+        kwargs.setdefault('editable', editable)
+        kwargs.setdefault('default', default)
+        super().__init__(*args, **kwargs)
