@@ -1,16 +1,27 @@
-from __future__ import unicode_literals
+from unittest import skip
 
-from unittest import skipUnless
-
-import django
+from django.core.cache import cache
 from django.core.exceptions import FieldError
 from django.test import TestCase
 
 from model_utils import FieldTracker
+from model_utils.tracker import DescriptorWrapper
 from tests.models import (
-    Tracked, TrackedFK, InheritedTrackedFK, TrackedNotDefault, TrackedNonFieldAttr, TrackedMultiple,
-    InheritedTracked, TrackedFileField,
-    ModelTracked, ModelTrackedFK, ModelTrackedNotDefault, ModelTrackedMultiple, InheritedModelTracked,
+    InheritedModelTracked,
+    InheritedTracked,
+    InheritedTrackedFK,
+    ModelTracked,
+    ModelTrackedFK,
+    ModelTrackedMultiple,
+    ModelTrackedNotDefault,
+    Tracked,
+    TrackedAbstract,
+    TrackedFileField,
+    TrackedFK,
+    TrackedMultiple,
+    TrackedNonFieldAttr,
+    TrackedNotDefault,
+    TrackerTimeStamped,
 )
 
 
@@ -46,7 +57,7 @@ class FieldTrackerTestCase(TestCase):
         self.instance.save()
 
 
-class FieldTrackerCommonTests(object):
+class FieldTrackerCommonTests:
 
     def test_pre_save_previous(self):
         self.assertPrevious(name=None, number=None)
@@ -74,7 +85,7 @@ class FieldTrackerTests(FieldTrackerTestCase, FieldTrackerCommonTests):
         self.assertChanged(name=None, number=None)
         self.instance.name = ''
         self.assertChanged(name=None, number=None)
-        self.instance.mutable = [1,2,3]
+        self.instance.mutable = [1, 2, 3]
         self.assertChanged(name=None, number=None, mutable=None)
 
     def test_pre_save_has_changed(self):
@@ -83,9 +94,14 @@ class FieldTrackerTests(FieldTrackerTestCase, FieldTrackerCommonTests):
         self.assertHasChanged(name=True, number=False, mutable=False)
         self.instance.number = 7
         self.assertHasChanged(name=True, number=True)
-        self.instance.mutable = [1,2,3]
+        self.instance.mutable = [1, 2, 3]
         self.assertHasChanged(name=True, number=True, mutable=True)
 
+    def test_save_with_args(self):
+        self.instance.number = 1
+        self.instance.save(False, False, None, None)
+        self.assertChanged()
+
     def test_first_save(self):
         self.assertHasChanged(name=True, number=False, mutable=False)
         self.assertPrevious(name=None, number=None, mutable=None)
@@ -93,22 +109,22 @@ class FieldTrackerTests(FieldTrackerTestCase, FieldTrackerCommonTests):
         self.assertChanged(name=None)
         self.instance.name = 'retro'
         self.instance.number = 4
-        self.instance.mutable = [1,2,3]
+        self.instance.mutable = [1, 2, 3]
         self.assertHasChanged(name=True, number=True, mutable=True)
         self.assertPrevious(name=None, number=None, mutable=None)
-        self.assertCurrent(name='retro', number=4, id=None, mutable=[1,2,3])
+        self.assertCurrent(name='retro', number=4, id=None, mutable=[1, 2, 3])
         self.assertChanged(name=None, number=None, mutable=None)
 
         self.instance.save(update_fields=[])
         self.assertHasChanged(name=True, number=True, mutable=True)
         self.assertPrevious(name=None, number=None, mutable=None)
-        self.assertCurrent(name='retro', number=4, id=None, mutable=[1,2,3])
+        self.assertCurrent(name='retro', number=4, id=None, mutable=[1, 2, 3])
         self.assertChanged(name=None, number=None, mutable=None)
         with self.assertRaises(ValueError):
             self.instance.save(update_fields=['number'])
 
     def test_post_save_has_changed(self):
-        self.update_instance(name='retro', number=4, mutable=[1,2,3])
+        self.update_instance(name='retro', number=4, mutable=[1, 2, 3])
         self.assertHasChanged(name=False, number=False, mutable=False)
         self.instance.name = 'new age'
         self.assertHasChanged(name=True, number=False)
@@ -120,14 +136,14 @@ class FieldTrackerTests(FieldTrackerTestCase, FieldTrackerCommonTests):
         self.assertHasChanged(name=False, number=True, mutable=True)
 
     def test_post_save_previous(self):
-        self.update_instance(name='retro', number=4, mutable=[1,2,3])
+        self.update_instance(name='retro', number=4, mutable=[1, 2, 3])
         self.instance.name = 'new age'
-        self.assertPrevious(name='retro', number=4, mutable=[1,2,3])
+        self.assertPrevious(name='retro', number=4, mutable=[1, 2, 3])
         self.instance.mutable[1] = 4
-        self.assertPrevious(name='retro', number=4, mutable=[1,2,3])
+        self.assertPrevious(name='retro', number=4, mutable=[1, 2, 3])
 
     def test_post_save_changed(self):
-        self.update_instance(name='retro', number=4, mutable=[1,2,3])
+        self.update_instance(name='retro', number=4, mutable=[1, 2, 3])
         self.assertChanged()
         self.instance.name = 'new age'
         self.assertChanged(name='retro')
@@ -136,8 +152,8 @@ class FieldTrackerTests(FieldTrackerTestCase, FieldTrackerCommonTests):
         self.instance.name = 'retro'
         self.assertChanged(number=4)
         self.instance.mutable[1] = 4
-        self.assertChanged(number=4, mutable=[1,2,3])
-        self.instance.mutable = [1,2,3]
+        self.assertChanged(number=4, mutable=[1, 2, 3])
+        self.instance.mutable = [1, 2, 3]
         self.assertChanged(number=4)
 
     def test_current(self):
@@ -146,29 +162,29 @@ class FieldTrackerTests(FieldTrackerTestCase, FieldTrackerCommonTests):
         self.assertCurrent(id=None, name='new age', number=None, mutable=None)
         self.instance.number = 8
         self.assertCurrent(id=None, name='new age', number=8, mutable=None)
-        self.instance.mutable = [1,2,3]
-        self.assertCurrent(id=None, name='new age', number=8, mutable=[1,2,3])
+        self.instance.mutable = [1, 2, 3]
+        self.assertCurrent(id=None, name='new age', number=8, mutable=[1, 2, 3])
         self.instance.mutable[1] = 4
-        self.assertCurrent(id=None, name='new age', number=8, mutable=[1,4,3])
+        self.assertCurrent(id=None, name='new age', number=8, mutable=[1, 4, 3])
         self.instance.save()
-        self.assertCurrent(id=self.instance.id, name='new age', number=8, mutable=[1,4,3])
+        self.assertCurrent(id=self.instance.id, name='new age', number=8, mutable=[1, 4, 3])
 
     def test_update_fields(self):
-        self.update_instance(name='retro', number=4, mutable=[1,2,3])
+        self.update_instance(name='retro', number=4, mutable=[1, 2, 3])
         self.assertChanged()
         self.instance.name = 'new age'
         self.instance.number = 8
-        self.instance.mutable = [4,5,6]
-        self.assertChanged(name='retro', number=4, mutable=[1,2,3])
+        self.instance.mutable = [4, 5, 6]
+        self.assertChanged(name='retro', number=4, mutable=[1, 2, 3])
         self.instance.save(update_fields=[])
-        self.assertChanged(name='retro', number=4, mutable=[1,2,3])
+        self.assertChanged(name='retro', number=4, mutable=[1, 2, 3])
         self.instance.save(update_fields=['name'])
         in_db = self.tracked_class.objects.get(id=self.instance.id)
         self.assertEqual(in_db.name, self.instance.name)
         self.assertNotEqual(in_db.number, self.instance.number)
-        self.assertChanged(number=4, mutable=[1,2,3])
+        self.assertChanged(number=4, mutable=[1, 2, 3])
         self.instance.save(update_fields=['number'])
-        self.assertChanged(mutable=[1,2,3])
+        self.assertChanged(mutable=[1, 2, 3])
         self.instance.save(update_fields=['mutable'])
         self.assertChanged()
         in_db = self.tracked_class.objects.get(id=self.instance.id)
@@ -176,24 +192,73 @@ class FieldTrackerTests(FieldTrackerTestCase, FieldTrackerCommonTests):
         self.assertEqual(in_db.number, self.instance.number)
         self.assertEqual(in_db.mutable, self.instance.mutable)
 
+    def test_refresh_from_db(self):
+        self.update_instance(name='retro', number=4, mutable=[1, 2, 3])
+        self.tracked_class.objects.filter(pk=self.instance.pk).update(
+            name='new age', number=8, mutable=[3, 2, 1])
+        self.assertChanged()
+        self.instance.name = 'like in db'
+        self.instance.number = 8
+        self.instance.mutable = [3, 2, 1]
+        self.assertChanged(name='retro', number=4, mutable=[1, 2, 3])
+        self.instance.refresh_from_db(fields=('name',))
+        self.assertChanged(number=4, mutable=[1, 2, 3])
+        self.instance.refresh_from_db(fields={'mutable'})
+        self.assertChanged(number=4)
+        self.instance.refresh_from_db()
+        self.assertChanged()
+
     def test_with_deferred(self):
         self.instance.name = 'new age'
         self.instance.number = 1
         self.instance.save()
-        item = list(self.tracked_class.objects.only('name').all())[0]
-        self.assertTrue(item._deferred_fields)
+        item = self.tracked_class.objects.only('name').first()
+        self.assertTrue(item.get_deferred_fields())
 
-        self.assertEqual(item.tracker.previous('number'), None)
-        self.assertTrue('number' in item._deferred_fields)
+        # has_changed() returns False for deferred fields, without un-deferring them.
+        # Use an if because ModelTracked doesn't support has_changed() in this case.
+        if self.tracked_class == Tracked:
+            self.assertFalse(item.tracker.has_changed('number'))
+            self.assertIsInstance(item.__class__.number, DescriptorWrapper)
+            self.assertTrue('number' in item.get_deferred_fields())
 
+        # previous() un-defers field and returns value
+        self.assertEqual(item.tracker.previous('number'), 1)
+        self.assertNotIn('number', item.get_deferred_fields())
+
+        # examining a deferred field un-defers it
+        item = self.tracked_class.objects.only('name').first()
         self.assertEqual(item.number, 1)
-        self.assertTrue('number' not in item._deferred_fields)
+        self.assertTrue('number' not in item.get_deferred_fields())
         self.assertEqual(item.tracker.previous('number'), 1)
         self.assertFalse(item.tracker.has_changed('number'))
 
+        # has_changed() returns correct values after deferred field is examined
+        self.assertFalse(item.tracker.has_changed('number'))
         item.number = 2
         self.assertTrue(item.tracker.has_changed('number'))
 
+        # previous() returns correct value after deferred field is examined
+        self.assertEqual(item.tracker.previous('number'), 1)
+
+        # assigning to a deferred field un-defers it
+        # Use an if because ModelTracked doesn't handle this case.
+        if self.tracked_class == Tracked:
+
+            item = self.tracked_class.objects.only('name').first()
+            item.number = 2
+
+            # previous() fetches correct value from database after deferred field is assigned
+            self.assertEqual(item.tracker.previous('number'), 1)
+
+            # database fetch of previous() value doesn't affect current value
+            self.assertEqual(item.number, 2)
+
+            # has_changed() returns correct values after deferred field is assigned
+            self.assertTrue(item.tracker.has_changed('number'))
+            item.number = 1
+            self.assertFalse(item.tracker.has_changed('number'))
+
 
 class FieldTrackerMultipleInstancesTests(TestCase):
 
@@ -377,7 +442,7 @@ class FieldTrackedModelMultiTests(FieldTrackerTestCase,
     def test_pre_save_previous(self):
         for tracker in self.trackers:
             self.tracker = tracker
-            super(FieldTrackedModelMultiTests, self).test_pre_save_previous()
+            super().test_pre_save_previous()
 
     def test_post_save_has_changed(self):
         self.update_instance(name='retro', number=4)
@@ -472,6 +537,30 @@ class FieldTrackerForeignKeyTests(FieldTrackerTestCase):
         self.assertCurrent(fk=self.instance.fk_id)
 
 
+class FieldTrackerTimeStampedTests(FieldTrackerTestCase):
+
+    fk_class = Tracked
+    tracked_class = TrackerTimeStamped
+
+    def setUp(self):
+        self.instance = self.tracked_class.objects.create(name='old', number=1)
+        self.tracker = self.instance.tracker
+
+    def test_set_modified_on_save(self):
+        old_modified = self.instance.modified
+        self.instance.name = 'new'
+        self.instance.save()
+        self.assertGreater(self.instance.modified, old_modified)
+        self.assertChanged()
+
+    def test_set_modified_on_save_update_fields(self):
+        old_modified = self.instance.modified
+        self.instance.name = 'new'
+        self.instance.save(update_fields=('name',))
+        self.assertGreater(self.instance.modified, old_modified)
+        self.assertChanged()
+
+
 class InheritedFieldTrackerTests(FieldTrackerTests):
 
     tracked_class = InheritedTracked
@@ -595,6 +684,16 @@ class ModelTrackerTests(FieldTrackerTests):
 
     tracked_class = ModelTracked
 
+    def test_cache_compatible(self):
+        cache.set('key', self.instance)
+        instance = cache.get('key')
+        instance.number = 1
+        instance.name = 'cached'
+        instance.save()
+        self.assertChanged()
+        instance.number = 2
+        self.assertHasChanged(number=True)
+
     def test_pre_save_changed(self):
         self.assertChanged()
         self.instance.name = 'new age'
@@ -603,7 +702,7 @@ class ModelTrackerTests(FieldTrackerTests):
         self.assertChanged()
         self.instance.name = ''
         self.assertChanged()
-        self.instance.mutable = [1,2,3]
+        self.instance.mutable = [1, 2, 3]
         self.assertChanged()
 
     def test_first_save(self):
@@ -613,16 +712,16 @@ class ModelTrackerTests(FieldTrackerTests):
         self.assertChanged()
         self.instance.name = 'retro'
         self.instance.number = 4
-        self.instance.mutable = [1,2,3]
+        self.instance.mutable = [1, 2, 3]
         self.assertHasChanged(name=True, number=True, mutable=True)
         self.assertPrevious(name=None, number=None, mutable=None)
-        self.assertCurrent(name='retro', number=4, id=None, mutable=[1,2,3])
+        self.assertCurrent(name='retro', number=4, id=None, mutable=[1, 2, 3])
         self.assertChanged()
 
         self.instance.save(update_fields=[])
         self.assertHasChanged(name=True, number=True, mutable=True)
         self.assertPrevious(name=None, number=None, mutable=None)
-        self.assertCurrent(name='retro', number=4, id=None, mutable=[1,2,3])
+        self.assertCurrent(name='retro', number=4, id=None, mutable=[1, 2, 3])
         self.assertChanged()
         with self.assertRaises(ValueError):
             self.instance.save(update_fields=['number'])
@@ -726,3 +825,9 @@ class InheritedModelTrackerTests(ModelTrackerTests):
         self.name2 = 'test'
         self.assertEqual(self.tracker.previous('name2'), None)
         self.assertTrue(self.tracker.has_changed('name2'))
+
+
+@skip
+class AbstractModelTrackerTests(ModelTrackerTests):
+
+    tracked_class = TrackedAbstract
